-- This file is a part of Falling Block Game 2.
-- 
-- Copyright (C) 2011 Jared Krinke.
-- 
-- Falling Block Game 2 is free software; you can redistribute it and/or
-- modify it under the terms of the GNU General Public License
-- as published by the Free Software Foundation; either version 2
-- of the License, or (at your option) any later version.
-- 
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
-- 
-- You should have received a copy of the GNU General Public License
-- along with this program; if not, write to the Free Software
-- Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

require("Fbg2/Common.lua")
require("Fbg2/PauseScreen.lua")
require("Effects/Ghost.lua")

local clearedToScore = { 40, 100, 300, 1200 }
local cleanupRows = 25

local donePeriod = 1000
local framePeriod = 16.64

local fastMoveInitialExtraDelay = 10 * framePeriod
local fastMovePeriod = 6 * framePeriod

local firstDropPeriod = 1000
local fastDropPeriod = 2 * framePeriod
local nextPieceDelay = 10 * framePeriod
local clearDelay = 20 * framePeriod
local scoreEffectPeriod = clearDelay * 1.4
local scoreEffectScaleMax = 1.5
local clearRowEffectPeriod = clearDelay
local clearRowEffectScaleMax = 2
local quadEffectPeriod = clearRowEffectPeriod
local quadEffectScaleMax = clearRowEffectScaleMax * 2
local cloudEffectScaleMin = 0.3
local cloudEffectScaleMax = 9
local cloudEffectPeriod = 150
local cloudEffectOffsetX = 0.6
local cloudEffectFloatUpOrigin = 0.08
local gameOverEffectPeriod = donePeriod / 2

local levelToDropPeriod = {
    [0] = 48 * framePeriod,
    [1] = 43 * framePeriod,
    [2] = 38 * framePeriod,
    [3] = 33 * framePeriod,
    [4] = 28 * framePeriod,
    [5] = 23 * framePeriod,
    [6] = 18 * framePeriod,
    [7] = 13 * framePeriod,
    [8] = 8 * framePeriod,
    [9] = 6 * framePeriod,
    [10] = 5 * framePeriod,
    [11] = 5 * framePeriod,
    [12] = 5 * framePeriod,
    [13] = 4 * framePeriod,
    [14] = 4 * framePeriod,
    [15] = 4 * framePeriod,
    [16] = 3 * framePeriod,
    [17] = 3 * framePeriod,
    [18] = 3 * framePeriod,
    [19] = 2 * framePeriod,
    [20] = 2 * framePeriod,
    [21] = 2 * framePeriod,
    [22] = 2 * framePeriod,
    [23] = 2 * framePeriod,
    [24] = 2 * framePeriod,
    [25] = 2 * framePeriod,
    [26] = 2 * framePeriod,
    [27] = 2 * framePeriod,
    [28] = 2 * framePeriod,
    [29] = 1 * framePeriod,
};

local initialHeightToRows = {
    [0] = 0,
    [1] = 3,
    [2] = 5,
    [3] = 8,
    [4] = 10,
    [5] = 12,
}

Board = { }

function Board.new(blockSize, scoreUpdated, ended)
    local board = Entity.new()

    board.overlayEffects = Entity.new(0, 0, 3)
    board.rowEffects = Entity.new(0, 0, 2)
    board.cloudEffects = Entity.new(0, 0, -0.25)

    board:addChild(board.overlayEffects)
    board:addChild(board.rowEffects)
    board:addChild(board.cloudEffects)

    board.pauseScreen = PauseScreen.new(board)
    board.scoreUpdated = scoreUpdated
    board.ended = ended
    board.score = 0
    board.cleared = 0
    board.level = 0
    board.nextPiece = Entity.new()
    board.nextPiece.width = blockSize
    board.nextPiece.height = blockSize
    board.controlPressed = {
        down = false,
        left = false,
        right = false,
    }
    board.config = {
        width = 10,
        height = 20,
    }
    board.config.rows = { }
    board.config.elements = { }
    board.pieces = {
        -- Left snake
        Piece.new(1, { { { 1, 0 }, { 2, 0 }, { 2, -1 }, { 3, -1 } },
                       { { 3, 1 }, { 2, 0 }, { 3, 0 },  { 2, -1 } } }),
        -- Right snake
        Piece.new(2, { { { 2, 0 }, { 3, 0 }, { 1, -1 }, { 2, -1 } },
                       { { 2, 1 }, { 2, 0 }, { 3, 0 },  { 3, -1 } } }),
        -- Left gun
        Piece.new(3, { { { 1, 0 }, { 2, 0 }, { 3, 0 },  { 3, -1 } },
                       { { 2, 1 }, { 2, 0 }, { 1, -1 }, { 2, -1 } },
                       { { 1, 1 }, { 1, 0 }, { 2, 0 },  { 3, 0 } },
                       { { 2, 1 }, { 3, 1 }, { 2, 0 },  { 2, -1 } }, }),
        -- Right gun
        Piece.new(4, { { { 1, 0 }, { 2, 0 }, { 3, 0 },  { 1, -1 } },
                       { { 1, 1 }, { 2, 1 }, { 2, 0 },  { 2, -1 } },
                       { { 3, 1 }, { 1, 0 }, { 2, 0 },  { 3, 0 } },
                       { { 2, 1 }, { 2, 0 }, { 2, -1 }, { 3, -1 } }, }),
        -- Square
        Piece.new(5, { { { 1, 0 }, { 2, 0 }, { 1, -1 },  { 2, -1 } } }),
        -- Line
        Piece.new(6, { { { 0, 0 }, { 1, 0 }, { 2, 0 },  { 3, 0 } },
                       { { 2, 2 }, { 2, 1 }, { 2, 0 },  { 2, -1 } }, }),
        -- Tee
        Piece.new(7, { { { 1, 0 }, { 2, 0 }, { 3, 0 }, { 2, -1 } },
                       { { 2, 1 }, { 1, 0 }, { 2, 0 }, { 2, -1 } },
                       { { 2, 1 }, { 1, 0 }, { 2, 0 }, { 3, 0 } },
                       { { 2, 1 }, { 2, 0 }, { 3, 0 }, { 2, -1 } }, }),
    }

    board.width = blockSize
    board.height = blockSize

    board.scoreEffect = Entity.new(board.config.width / 2, 0.5, 1, 0.75, 1.5)
    board.scoreEffect.element = board.scoreEffect.elements:add(Element.Text.new("", 0, -0.5, 0, 1, 1, 0, nil, "center"))
    board.rowEffect = Entity.new(board.config.width / 2, 0.5)
    board.quadEffect = Entity.new(board.config.width / 2, 0.5, 0.5, board.config.width, 4)
    board.quadEffect.element = board.quadEffect.elements:add(Element.Image.new("Images/Blank.png"))
    board.quadEffect.element.color = Color.new(1, 1, 1, 0.75)
    board.cloudEffect = Entity.new()
    board.cloudEffect.color = Color.new(0.4, 0.4, 0.4)
    board.cloudEffect.element = board.cloudEffect.elements:add(Element.Image.new("Images/Cloud.png", 0, 0.08, 0, cloudEffectScaleMin, cloudEffectScaleMin))
    board.gameOverEffect = Entity.new(board.config.width / 2, board.config.height / 2, 0, board.config.width * 1.4, board.config.width * 0.7)
    board.gameOverEffect.elements:add(Element.Image.new("Images/GameOver.png"))
    board.youWinEffect = Entity.new(board.config.width / 2, board.config.height / 2, 0, board.config.width * 1.4, board.config.width * 0.7)
    board.youWinEffect.elements:add(Element.Image.new("Images/YouWin.png"))

    local j = 1

    while j <= board.config.height do
        -- Group all elements in a row into a single entity for convenience
        local rowEntity = Entity.new(0, j - 1)
        local i = 1

        board.config[j] = { }
        board.config.elements[j] = { }

        while i <= board.config.width do
            local element = Element.Image.new("Images/Blank.png", i - 0.5, 0.5)

            -- Default to no block
            element.color = Color.new(1, 1, 1, 0)
            board.config[j][i] = 0
            board.config.elements[j][i] = rowEntity.elements:add(element)
            i = i + 1
        end

        board:addChild(rowEntity)
        board.config.rows[j] = rowEntity
        j = j + 1
    end

    local i = 1
    board.rowEffect.config = { }

    while i <= board.config.width do
        local element = Element.Image.new("Images/Blank.png", i - 0.5 - board.config.width / 2, 0)
        
        board.rowEffect.config[i] = board.rowEffect.elements:add(element)
        i = i + 1
    end

    function board.setBlock(self, i, j, value)
        local config = board.config

        if value ~= config[j][i] and value >= 0 and value < #Fbg2.blockImages then
            local elements = board.config.elements
            
            -- Update the value at this position
            config[j][i] = value
            
            -- Update the element at this position
            if value ~= 0 then
                elements[j][i].image = Fbg2.blockImages[value + 1]
                elements[j][i].color.opacity = 1
            else
                elements[j][i].color.opacity = 0
            end
        end
    end

    function board.clear(self)
        local config = board.config
        local j = 1

        while j <= config.height do
            local i = 1

            while i <= config.width do
                self:setBlock(i, j, 0)
                i = i + 1
            end

            j = j + 1
        end
    end

    function board.addGarbage(self, rows)
        local config = board.config
        local j = 1

        while j <= rows and j <= config.height do
            local valid = false
            local i = 1

            while i <= config.width do
                -- 45 % of spaces are occupied
                if random() <= 0.45 then
                    self:setBlock(i, j, random(#self.pieces) + 1)
                else
                    self:setBlock(i, j, 0)
                    valid = true
                end

                i = i + 1
            end

            -- Re-do this row if it is completely full
            if valid then
                j = j + 1
            end
        end
    end

    function board.getBlockOccupied(self, i, j)
        if isNil(self.config[j][i]) then debugBreak() end
        return self.config[j][i] ~= 0
    end

    function board.getWidth(self)
        return self.config.width
    end

    function board.getHeight(self)
        return self.config.height
    end

    function board.getNextPiece(self, x, y)
        self.nextPiece.x = x
        self.nextPiece.y = y

        return self.nextPiece
    end

    function board.hidePiece(self)
        if not isNil(self.piece) then
            self:removeChild(self.piece)
            self.piece = nil
        end
    end

    function board.chooseNextPiece(self)
        local index = random(#self.pieces) + 1
        local nextPiece = self.pieces[index]

        self.nextPiece.index = index
        self.nextPiece.elements = nextPiece.configs[1].elements
    end

    function board.showNewPiece(self)
        self:hidePiece()

        if isNil(self.nextPiece.index) then
            self:chooseNextPiece()
        end

        self.piece = self.pieces[self.nextPiece.index]
        self:chooseNextPiece()
        self.piece:reset(self, 4, 20)
        self:addChild(self.piece)
    end

    -- TODO: Consider making this more efficient by just reassigning rows
    function board.removeRow(self, row)
        local config = board.config
        local width = config.width
        local height = config.height
        local j = row

        while j <= height do
            local j0 = j + 1
            local i = 1

            while i <= width do
                local value = 0

                if j0 <= height then
                    value = config[j0][i]
                end

                self:setBlock(i, j, value)
                i = i + 1
            end

            j = j + 1
        end
    end

    function board.resetOverlayEffects(self)
        self.overlayEffects:clearChildren()
    end
    
    function board.resetGameplayEffects(self)
        self.rowEffects:clearChildren()
        self.cloudEffects:clearChildren()

        -- Reset row positions
        local config = board.config
        local height = config.height
        local j = 1

        while j <= height do
            config.rows[j].y = j - 1
            j = j + 1
        end
    end

    -- TODO: Special effect for clearing 4 rows at once?
    function board.clearRowEffect(self, currentRow, cleared)
        local config = board.config
        local width = config.width
        local height = config.height
        local i = 1
        local row = currentRow + cleared

        -- Create ghost effect
        self.rowEffect.y = row - 0.5

        while i <= width do
            local value = config[row][i]

            self.rowEffect.config[i].image = Fbg2.blockImages[value + 1]
            i = i + 1
        end

        Ghost.new(self.rowEffect, clearRowEffectPeriod, clearRowEffectScaleMax):setParent(self.rowEffects)

        -- Create special quad effect, if this is for the fourth row
        if cleared == 3 then
            self.quadEffect.y = row - 2
            Ghost.new(self.quadEffect, quadEffectPeriod, quadEffectScaleMax):setParent(self.rowEffects)
        end

        -- Create entity to slide row down
        local slider = Entity.new()

        slider.timer = 0
        slider.offset = 0

        function slider.update(s, ms)
            local lastOffset = s.offset

            s.timer = s.timer + ms
            s.offset = max(0, 1 - s.timer / clearDelay)

            -- Adjust each row down
            local delta = s.offset - lastOffset
            local j = currentRow

            while j <= height do
                local rowEntity = config.rows[j]

                -- Adjust position
                rowEntity.y = rowEntity.y + delta
                j = j + 1
            end

            -- Remove once effect is complete
            if s.timer >= clearDelay then
                self:resetGameplayEffects()
            end
        end

        self.rowEffects:addChild(slider)
        slider:update(0)
    end

    function board.clearCompletedRows(self)
        local config = board.config
        local width = config.width
        local height = config.height
        local j = 1
        local cleared = 0
        local top = 0

        while j + cleared <= height do
            local completed = true
            local i = 1

            -- Check to see if the entire row is occupied
            while completed and i <= width do
                if not self:getBlockOccupied(i, j) then
                    completed = false
                end

                i = i + 1
            end

            -- Remove the row, if necessary
            if completed then
                self:clearRowEffect(j, cleared)
                self:removeRow(j)
                top = j + cleared
                cleared = cleared + 1
            else
                j = j + 1
            end
        end

        return cleared, top
    end

    function board.startFastDrop(self, row)
        self.fastDrop = true
        self.fastDropStart = row
    end

    function board.stopFastDrop(self)
        self.fastDrop = false
    end

    function board.resetControlPressed(self)
        self.controlPressed.down = false
        self.controlPressed.left = false
        self.controlPressed.right = false
    end

    function board.reset(self, mode, level, initialHeight)
        self.score = 0
        self.mode = mode
        self.level = level
        self.initialHeight = initialHeight
        self.dropPeriod = firstDropPeriod

        -- Cleanup mode uses a countdown to 0 rows
        if self.mode == Fbg2.modes.Cleanup then
            self.cleared = cleanupRows
        else
            self.cleared = 0
        end
        
        if isFunction(self.scoreUpdated) then
            self.scoreUpdated(self.score, self.cleared, self.level, 0)
        end

        self:clear()
        self.done = false
        self.paused = false
        self.lost = false
        self.dropElapsed = 0
        self.fastMoveElapsed = 0
        self.firstDrop = true
        self:stopFastDrop()
        self:showNewPiece()

        -- Reset control states
        self:resetControlPressed()

        -- Reset effects
        self:resetOverlayEffects()
        self:resetGameplayEffects()

        -- Cleanup mode can have garbage
        if self.mode == Fbg2.modes.Cleanup then
            local rows = initialHeightToRows[self.initialHeight]

            if isNumber(rows) then
                self:addGarbage(rows)
            end
        end
    end

    function board.updateScore(self, cleared)
        local originalScore = self.score
        local fastDropPoints = 0

        -- Fast drop yields one point per row
        if self.fastDrop then
            local i, j = self.piece:getPosition()

            fastDropPoints = self.fastDropStart - j
            self.score = self.score + fastDropPoints
        end

        -- Line scores multiply by level and depend on the number of rows cleared
        if cleared >= 1 and cleared <= #clearedToScore then
            self.score = self.score + clearedToScore[cleared] * (self.level + 1)

            if self.mode == Fbg2.modes.Cleanup then
                -- Cleanup mode counts down
                self.cleared = max(0, self.cleared - cleared)

                -- Check for end
                if self.cleared <= 0 then
                    self:endGame(false)
                end
            else
                -- Other modes count up and advance levels
                -- Update level
                self.cleared = self.cleared + cleared

                if self.level < floor(self.cleared / 10) then
                    self.level = self.level + 1
                end
            end
        end

        -- Notify handler that score has been updated
        if originalScore ~= self.score and isFunction(self.scoreUpdated) then
            self.scoreUpdated(self.score, self.cleared, self.level, cleared)
        end

        -- Return the total points and points derived from lines (displayed on the board)
        local points = self.score - originalScore

        return points, points - fastDropPoints
    end

    function board.getScore(self)
        return self.score, self.cleared, self.level
    end

    function board.moveLeft(self)
        if not isNil(self.piece) then
            if self.piece:tryMoveLeft(self) then
                Audio.play("Sounds/Move.wav", 0.5)
            end
        end
    end

    function board.moveRight(self)
        if not isNil(self.piece) then
            if self.piece:tryMoveRight(self) then
                Audio.play("Sounds/Move.wav", 0.5)
            end
        end
    end

    function board.moveDown(self)
        if not isNil(self.piece) then
            if self.piece:tryMoveDown(self) then
                -- Check to see if the piece should slide underneath another one
                if self.controlPressed.left then
                    local i, j = self.piece:getPosition()

                    if not self.piece:validate(self, nil, i - 1, j + 1)
                        and self.piece:validate(self, nil, i - 1, j)
                    then
                        -- Piece couldn't be moved before, but now can
                        self:moveLeft()
                    end
                end

                if self.controlPressed.right then
                    local i, j = self.piece:getPosition()

                    if not self.piece:validate(self, nil, i + 1, j + 1)
                        and self.piece:validate(self, nil, i + 1, j)
                    then
                        self:moveRight()
                    end
                end
            else
                -- Piece can't move down, so complete
                self.piece:complete(self)

                -- Add cloud effects where necessary
                self.piece:processCloudEffects(self, function(i, j, direction, floatUp)
                    -- Move generic cloud effect
                    local cloudEffect = self.cloudEffect
                    local x = i - 1

                    if direction > 0 then
                        x = x + 1
                    end

                    cloudEffect.x = x
                    cloudEffect.y = j - 1

                    -- Float up if there's a block below this space
                    if floatUp then
                        cloudEffect.element.y = cloudEffectFloatUpOrigin
                    else
                        cloudEffect.element.y = 0
                    end

                    -- Actually create the effect
                    Ghost.new(cloudEffect, cloudEffectPeriod, cloudEffectScaleMax, true, nil, direction * cloudEffectOffsetX):setParent(self.cloudEffects)
                end)

                -- Check for completed rows
                local cleared, top = self:clearCompletedRows()
                local points, linePoints = self:updateScore(cleared)

                -- Add score effect when a line is completed
                if cleared > 0 then
                    self.scoreEffect.y = top - 0.5
                    self.scoreEffect.element.text = String.format("+%d", linePoints)
                    Ghost.new(self.scoreEffect, scoreEffectPeriod, scoreEffectScaleMax):setParent(self.rowEffects)
                end

                -- Reset fast drop
                self:stopFastDrop()

                -- Remove piece and set timer for next piece
                self:hidePiece()
                self.nextPieceTimer = -nextPieceDelay

                if cleared > 0 then
                    self.nextPieceTimer = self.nextPieceTimer - clearDelay

                    if cleared >= 4 then
                        Audio.play("Sounds/Quadruple.wav", 0.8)
                    else
                        Audio.play("Sounds/Clear.wav", 0.75)
                    end
                else
                    Audio.play("Sounds/Land.wav")
                end
            end
        end
    end

    function board.rotateClockwise(self)
        if not isNil(self.piece) then
            if self.piece:tryRotateClockwise(self) then
                Audio.play("Sounds/Rotate.wav", 0.5)
            end
        end
    end

    function board.rotateCounterclockwise(self)
        if not isNil(self.piece) then
            if self.piece:tryRotateCounterclockwise(self) then
                Audio.play("Sounds/Rotate.wav", 0.5)
            end
        end
    end

    function board.getLevelDropPeriod(self)
        local period = levelToDropPeriod[self.level]

        -- Level 29 is as fast as it gets
        if self.level > #levelToDropPeriod then
            period = levelToDropPeriod[#levelToDropPeriod]
        end

        return period
    end

    function board.getDropPeriod(self)
        return self.dropPeriod
    end

    function board.endGameEffect(self, lost)
        local base = self.gameOverEffect
        local sound = "Sounds/Boom.wav"

        if not lost then
            base = self.youWinEffect
            sound = "Sounds/Quadruple.wav"
        end

        local effect = ScriptedEntity.new(base.elements,
                                          { { 0, base.x, base.y, 0, base.width / 100, base.height / 100, 0, Color.new(1, 1, 1, 0) },
                                            { gameOverEffectPeriod, base.x, base.y, 0, base.width, base.height, 360, Color.new(1, 1, 1, 1) } },
                                          false)

        self.overlayEffects:addChild(effect)
        Audio.play(sound)
    end

    function board.endGame(self, lost, external)
        self.paused = true
        self.done = false
        self.doneElapsed = 0

        if isBoolean(lost) then
            self.lost = lost
        end

        -- Only add effects when the end was triggered normally
        if not isBoolean(external) or not external then
            self:endGameEffect(self.lost)
        end
    end

    function board.update(self, ms)
        if not self.paused then
            -- Check for next piece timer
            if isNil(self.piece) and self.nextPieceTimer < 0 then
                self.nextPieceTimer = self.nextPieceTimer + ms
            else
                -- Next piece timer just fired, so add a new piece
                if isNil(self.piece) then
                    self:showNewPiece()

                    -- Check for loss
                    if not isNil(self.piece) and not self.piece:validate(self) then
                        self:endGame(true)
                    end
                end

                -- Game may be paused at this point due to a loss
                if not self.paused then
                    -- Check for fast move
                    if self.controlPressed.left or self.controlPressed.right then
                        self.fastMoveElapsed = self.fastMoveElapsed + ms

                        -- Check for first fast move (which has an extra delay)
                        while self.fastMoveElapsed >= fastMovePeriod do
                            self.fastMoveElapsed = self.fastMoveElapsed - fastMovePeriod

                            if self.controlPressed.left then
                                self:moveLeft()
                            elseif self.controlPressed.right then
                                self:moveRight()
                            end
                        end
                    end

                    -- Determine drop period
                    self.dropPeriod = self:getLevelDropPeriod()

                    -- First drop is slower than normal
                    if self.firstDrop then
                        self.dropPeriod = firstDropPeriod
                    end

                    -- Check for fast drop
                    if self.fastDrop then
                        self.dropPeriod = min(fastDropPeriod, self.dropPeriod)
                    end

                    -- Check for drop
                    self.dropElapsed = self.dropElapsed + ms

                    while self.dropElapsed >= self.dropPeriod do
                        self:moveDown()
                        self.dropElapsed = self.dropElapsed - self.dropPeriod
                        self.firstDrop = false
                    end
                end
            end
        elseif not self.done then
            -- Wait a short while before allowing the player to exit
            self.doneElapsed = self.doneElapsed + ms

            if self.doneElapsed >= donePeriod then
                self.done = true
            end
        end

        self:updateChildren(ms)
    end

    function board.exit(self)
        if isFunction(self.ended) then
            -- Losing will not save the high score on Cleanup mode
            local valid = (not self.lost or self.mode ~= Fbg2.modes.Cleanup)

            self.ended(valid, self.mode, self.initialHeight, self:getScore())
        end
    end

    local eventHandlers = {
        left = function(self, pressed)
            if not self.paused then
                self.controlPressed.left = pressed

                if pressed then
                    self:moveLeft()

                    -- Prepare for fast move
                    self.fastMoveElapsed = -fastMoveInitialExtraDelay
                end
            end
        end,

        right = function(self, pressed)
            if not self.paused then
                self.controlPressed.right = pressed

                if pressed then
                    self:moveRight()

                    -- Prepare for fast move
                    self.fastMoveElapsed = -fastMoveInitialExtraDelay
                end
            end
        end,

        down = function(self, pressed)
            if not self.paused then
                self.controlPressed.down = pressed

                -- Reset elapsed time at this level when enabling fast drop
                if not isNil(self.piece) and pressed then
                    local i, j = self.piece:getPosition()
                    self:startFastDrop(j)

                    -- Don't drop multiple times due to accumulated time at this level
                    self.dropElapsed = min(self.dropElapsed, fastDropPeriod, self:getLevelDropPeriod())
                else
                    self:stopFastDrop()
                end
            end
        end,

        rotateLeft = function(self, pressed)
            if not self.paused then
                if pressed then
                    self:rotateCounterclockwise()
                end
            end
        end,

        rotateRight = function(self, pressed)
            if not self.paused then
                if pressed then
                    self:rotateClockwise()
                end
            end
        end,

        cancel = function(self, pressed)
            if pressed then
                if self.paused then
                    -- Override delay before allowing exit
                    self:exit()
                else
                    self.pauseScreen:show()
                end
            end
        end,
    }

    eventHandlers.up = eventHandlers.rotateRight
    eventHandlers.activate = eventHandlers.rotateLeft

    function board.inputReceived(self, event, pressed)
        local eventHandler = eventHandlers[event]
        local handled = false

        if isFunction(eventHandler) then
            eventHandler(self, pressed)
            handled = true
        end

        -- Any input will exit once the "done" flag is set
        if self.done and not pressed then
            self:exit()
        end

        return handled
    end

    return board
end

